/*
Copyright © 2013 Alexander Kobzar (sashakobzar@gmail.com)

GETALP
Grenoble Informatics Laboratory
France

This scripts allows you to extract sentences containing multi-word expressions (MWEs) identified by mwetoolkit (candidates.xml).
It splits both source and target texts into two parts: one containing only sentences with MWEs and another one without MWEs. 
This data can be used to train and test a machine translation system.
It also extracts data required by the mwe2blast script in order to prepare a human evaluation experiment (Blast annotation file). 
Moreover, it can extract a list of MWEs accompanied by correspodning sentences and statistics on the frequency of each MWEs (lemma and surface form) and on the number of MWEs based on their frequency.
*/
using System;
using System.IO;
using System.Xml;
using System.Collections.Generic;
using System.Linq;
using System.Text;

class ExtractSent
{
    //Close StreamWriters
    public static void CloseWriters(params StreamWriter[] writers)
    {
        foreach (StreamWriter stWriter in writers)
            stWriter.Close();
    }

    public static void Main(string[] args)
    {
        bool moses = false, blast = false, MWESent = false, freq = false;
        int dev = 0, totalMWENumber = 0, frequency = 0, sentenceIndex, corpusLineCount = 0;
        string sourceFile, targetFile, candFile, surface, lemma, sourceLine, targetLine;
        string[] surfWords;
        Dictionary<int, List<int[]>> sentenceMWEPos = new Dictionary<int, List<int[]>>();
        SortedDictionary<int, int> MWEFreqOfAppear = new SortedDictionary<int, int>();
        Dictionary<int, List<string[]>> sentenceMWE = new Dictionary<int, List<string[]>>();
        Dictionary<string, int> MWESurfaceFrequency = new Dictionary<string, int>(), MWELemmaFrequency = new Dictionary<string, int>();
        StreamReader sourceCorpus, targetCorpus;
        XmlTextReader candidatesReader;
        StreamWriter sourceTrainingCorpus = null, sourceTestCorpus = null, targetTestCorpus = null, targetTrainingCorpus = null,
        MWEBlast = null, MWESentWriter = null, freqOfAppearWriter = null, surfaceFreqWriter = null, lemmaFreqWriter = null;
        string message = @"splitcorpus.exe <source text file> <target text file> <mwetoolkit candidates file> [OPTIONS]

Options:
--------
-moses - splits source and target texts. Four files are generated with their names having either 'train' or 'test' prefix followed by a correspodning file name

-blast - generates mwe2blast.txt required by mwe2blast

-dev <numberoflines> - use this option if you extracted a part of text from the beggining of each corpus file (development set) after you identified MWEs with mwetoolkit

-mwesent - generates mwe-sentence.txt containing MWEs with corresponding sentences. Both surface forms and lemmas are specified. All data is tab-separated

-freq - generates three files: freqofappear.txt, surffreq.txt, and lemmafreq.txt. The first one contains statistics on the number of MWEs based on their frequency. Others contain frequency information on the surface form and lemma of each MWE.  All data is tab-separated";
        if (args.Count() != 0 && (args[0] == "/?" || args[0] == "-help"))
        {
            Console.WriteLine(message);
            return;
        }
        if (args.Count() < 3)
        {
            Console.WriteLine("You should specify all required parameters and at least one option except -dev.\r\nType '[mono] filterblast.exe -help' or '[mono] filterblast.exe /?' for help.");
            return;
        }
        #region Process command-line arguments
        //Check whether specified files exist
        for (int i = 0; i <= 2; i++)
        {
            if (!File.Exists(args[i]))
            {
                Console.WriteLine("{0} does not exist", args[i]);
                return;
            }
        }
        //Check command-line options
        for (int i = 3; i < args.Count(); i++)
        {
            switch (args[i])
            {
                case "-moses":
                    moses = true;
                    break;
                case "-blast":
                    blast = true;
                    break;
                case "-mwesent":
                    MWESent = true;
                    break;
                case "-freq":
                    freq = true;
                    break;
                case "-dev":
                    if (!int.TryParse(args[Array.IndexOf(args, "-dev") + 1], out dev))
                    {
                        Console.WriteLine("-dev is specified incorrectly. Please, provide an integer.");
                        return;
                    }
                    i++;
                    break;
                default:
                    Console.WriteLine((args[i].StartsWith("-") ? "Invalid option: " : "Invalid argument: ") + args[i] + "\r\nType '[mono] filterblast.exe -help' or '[mono] filterblast.exe /?' for help.");
                    return;
            }
        }
        if (!(moses || blast || MWESent || freq))
        {
            Console.WriteLine("You should specify at least one option except -dev.\r\nType '[mono] filterblast.exe -help' or '[mono] filterblast.exe /?' for help.");
            return;
        }
        #endregion
        sourceFile = args[0];
        targetFile = args[1];
        candFile = args[2];
        sourceCorpus = new StreamReader(sourceFile);
        targetCorpus = new StreamReader(targetFile);
        candidatesReader = new XmlTextReader(candFile);
        /* A candidates xml file contains a reference to a DTD file. If this file is missing, Microsoft CLR throws an exception whereas Mono does not.
         * There is no DtdProcessing enumeration in the XmlTextReader class in the version of Mono used to develop this script - version 2.10.8.1 on Ubuntu.
         * Moreover, if the XmlResolver property is set to null then an exception is thrown as well in this version of Mono.
         * The following code handles this problem.*/
		try
        {
            candidatesReader.ReadToFollowing("occurs");
        }
        catch
        {
            candidatesReader.XmlResolver = null;
            candidatesReader.ReadToFollowing("occurs");
        }
        //Extract word positions of MWE in a sentence, its frequency, and IDs of sentences containing MWEs
        do
        {
            //Extract lemmas and surface forms of MWE 
            lemma = "";
            surface = "";
            #region Check whether an XML file contains required information
            if (!candidatesReader.ReadToDescendant("w"))
            {
                Console.WriteLine("{0} is malformed.", candFile);
                candidatesReader.Close();
                return;
            }
            while (candidatesReader.Name == "w")
            {
                if (candidatesReader.GetAttribute("lemma") == null || candidatesReader.GetAttribute("surface") == null)
                {
                    Console.WriteLine("{0} is malformed.", candFile);
                    candidatesReader.Close();
                    return;
                }
                lemma += candidatesReader["lemma"].ToLower() + " ";
                surface += candidatesReader["surface"].ToLower() + " ";
                candidatesReader.ReadStartElement(candidatesReader.Name);
                candidatesReader.Read();
            }
            if (!candidatesReader.ReadToFollowing("sources"))
            {
                Console.WriteLine("{0} is malformed.", candFile);
                candidatesReader.Close();
                return;
            }
            if (candidatesReader.GetAttribute("ids") == null)
            {
                Console.WriteLine("{0} is malformed.", candFile);
                candidatesReader.Close();
                return;
            }
            #endregion
            //Enumerate word positions				
            foreach (string s in candidatesReader["ids"].Contains(";") ? candidatesReader["ids"].Split(new char[] { ';' }) : new string[] { candidatesReader["ids"] })
            {
                sentenceIndex = Convert.ToInt32(s.Substring(0, s.IndexOf(":"))) - dev;
                //Frequency disregarding the development set 
                if (Convert.ToInt32(s.Substring(0, s.IndexOf(":"))) > dev)
                {
                    frequency++;
                    //Word and sentence positions
                    if (!sentenceMWEPos.ContainsKey(sentenceIndex))
                        sentenceMWEPos.Add(sentenceIndex, new List<int[]>{Array.ConvertAll <string,int> (s.Substring (s.IndexOf (':') + 1).Split (new char []{','}), delegate(string dstr){
							return int.Parse (dstr) - 1;}).ToArray ()}
                        );
                    else
                        sentenceMWEPos[sentenceIndex].Add(Array.ConvertAll<string, int>(s.Substring(s.IndexOf(':') + 1).Split(new char[] { ',' }), delegate(string dstr)
                        {
                            return Convert.ToInt32(dstr) - 1;
                        }
                        ).ToArray()
                        );
                }
                else
                    continue;
                //Sentence position and MWE				
                if (!sentenceMWE.ContainsKey(sentenceIndex))
                    sentenceMWE.Add(sentenceIndex, new List<string[]> { lemma.TrimEnd().Split(new char[] { ' ' }) });
                else
                    sentenceMWE[sentenceIndex].Add(lemma.TrimEnd().Split(new char[] { ' ' }));
            }
            //Update frequency information	
            if ((frequency != 0))
            {
                if (!MWELemmaFrequency.ContainsKey(lemma))
                    MWELemmaFrequency.Add(lemma, frequency);
                else
                    MWELemmaFrequency[lemma] += frequency;
                if (!MWESurfaceFrequency.ContainsKey(surface))
                    MWESurfaceFrequency.Add(surface, frequency);
                else
                    MWESurfaceFrequency[surface] += frequency;
                totalMWENumber++;
                if (!MWEFreqOfAppear.ContainsKey(frequency))
                    MWEFreqOfAppear.Add(frequency, 1);
                else
                    MWEFreqOfAppear[frequency]++;
                frequency = 0;
            }
        }
        while (candidatesReader.ReadToFollowing("occurs"));
        candidatesReader.Close();
        //Create files for the training and test sets
        if (moses)
        {
            sourceTestCorpus = new StreamWriter(File.Open("test." + sourceFile, FileMode.Create));
            sourceTrainingCorpus = new StreamWriter(File.Open("train." + sourceFile, FileMode.Create));
            targetTestCorpus = new StreamWriter(File.Open("test." + targetFile, FileMode.Create));
            targetTrainingCorpus = new StreamWriter(File.Open("train." + targetFile, FileMode.Create));
        }
        //Create a file which will contain MWEs with corresponding sentences
        if (MWESent)
        {
            MWESentWriter = new StreamWriter("mwe-sentence.txt");
            MWESentWriter.WriteLine("Surface form\tLemma\tSentence");
        }
        //Create a file for the preparation of human evaluation
        if (blast)
        {
            MWEBlast = new StreamWriter(File.Open("mwe2blast.txt", FileMode.Create));
        }
        sentenceIndex = 1;
        //Write information to files
        while ((sourceLine = sourceCorpus.ReadLine()) != null)
        {
            corpusLineCount++;
            if ((targetLine = targetCorpus.ReadLine()) == null)
            {
                Console.WriteLine("{0} has more sentences than {1}.", sourceFile, targetFile);
                sourceCorpus.Close();
                targetCorpus.Close();
                return;
            }
            if (sentenceMWEPos.ContainsKey(sentenceIndex))
            {
                //Test sets
                if (moses)
                {
                    sourceTestCorpus.WriteLine(sourceLine);
                    targetTestCorpus.WriteLine(targetLine);
                }
                surfWords = sourceLine.Split(new char[] { ' ' });
                foreach (int[] expr in sentenceMWEPos[sentenceIndex])
                {
                    foreach (int i in expr)
                    {
                        {
                            if (i >= surfWords.Count())
                            {
                                Console.WriteLine("Sentence (line) {0} in {1} does not correspond to its reference in {2}. " + (dev == 0 ? "If you extracted a development set from your source file, you should use the -dev option to specify the number of sentences removed as development set from {1}." : "Probably, the -dev value you specified, {3}, is not equal to the number of removed sentences from {1}."), sentenceIndex, sourceFile, candFile, dev);
                                sourceCorpus.Close();
                                targetCorpus.Close();
                                return;
                            }
                        }
                    }
                }
                if ((blast) || (MWESent))
                {
                    for (int list = 0; list < sentenceMWEPos[sentenceIndex].Count(); list++)
                    {
                        //MWE's positions
                        if (blast)
                        {
                            MWEBlast.Write(string.Join(" ", sentenceMWEPos[sentenceIndex][list]) + "\t");
                        }//MWE accompanied by its sentence
                        if (MWESent)
                        {
                            foreach (int i in sentenceMWEPos[sentenceIndex][list])
                            {
                                surfWords[i] = surfWords[i].ToUpper();
                                MWESentWriter.Write(surfWords[i] + " ");
                            }
                            MWESentWriter.Write("\t" + string.Join(" ", sentenceMWE[sentenceIndex][list]).ToUpper());
                            MWESentWriter.WriteLine("\t" + string.Join(" ", surfWords));
                            surfWords = sourceLine.Split(new char[] { ' ' });
                        }
                    }
                }
                //Source sentence
                if (blast)
                {
                    MWEBlast.WriteLine();
                    MWEBlast.WriteLine(sourceLine);
                }
            }
            else
            {
                //Training sets
                if (moses)
                {
                    sourceTrainingCorpus.WriteLine(sourceLine);
                    targetTrainingCorpus.WriteLine(targetLine);
                }
            }
            sentenceIndex++;
        }
        if (targetCorpus.Peek() != -1)
        {
            Console.WriteLine("{0} has more sentences than {1}.", targetFile, sourceFile);
            sourceCorpus.Close();
            targetCorpus.Close();
            return;
        }
        //Write the frequency information to files
        if (freq)
        {
            freqOfAppearWriter = new StreamWriter("freqofappear.txt");
            freqOfAppearWriter.WriteLine("Frequency\tNumber of MWEs\tPercentage");
            surfaceFreqWriter = new StreamWriter("surffreq.txt");
            surfaceFreqWriter.WriteLine("Surface\tFrequency");
            lemmaFreqWriter = new StreamWriter("lemmafreq.txt");
            lemmaFreqWriter.WriteLine("Lemma\tFrequency");
            foreach (int i in MWEFreqOfAppear.Keys)
            {
                freqOfAppearWriter.WriteLine(i.ToString() + "\t" + MWEFreqOfAppear[i] + "\t" + MWEFreqOfAppear[i] * 100.0 / totalMWENumber + "%");
            }
            foreach (KeyValuePair<string, int> kvp in MWESurfaceFrequency.OrderBy(x => x.Key).OrderByDescending(x => x.Value))
            {
                surfaceFreqWriter.WriteLine(kvp.Key + "\t" + kvp.Value);
            }
            foreach (KeyValuePair<string, int> kvp in MWELemmaFrequency.OrderBy(x => x.Key).OrderByDescending(x => x.Value))
            {
                lemmaFreqWriter.WriteLine(kvp.Key + "\t" + kvp.Value);
            }
            CloseWriters(freqOfAppearWriter, surfaceFreqWriter, lemmaFreqWriter);
        }
        if (moses)
        {
            CloseWriters(sourceTestCorpus, sourceTrainingCorpus, targetTestCorpus, targetTrainingCorpus);
        }
        if (blast)
            MWEBlast.Close();
        if (MWESent)
            MWESentWriter.Close();
        Console.WriteLine("Surface forms: " + MWESurfaceFrequency.Count);
        Console.WriteLine("Lemmas: " + MWELemmaFrequency.Count);
        if (moses)
        {
            Console.WriteLine("Number of lines in train corpus: " + (corpusLineCount - sentenceMWEPos.Count));
            Console.WriteLine("Number of lines in test corpus (to be included into Blast file): " + sentenceMWEPos.Count);
        }
    }
}